# -*- coding: utf-8 -*-
import heimdall
import pytest
from sys import version_info as _py

"""
Some of these tests expect a specific MariaDB or MySQL setup:
root user present with password 'password', port 3306.
In addition to these config parameters (which are smoke-tested
by unit test `test_smoke_url_format`), a DB called `menagerie`
must exist, with contents described by the SQL dump test file
MENAGERIE_BASIC_SQL, defined in `/test/stubs.py` .
"""


@pytest.skip(reason="TODO: IC with mariadb", allow_module_level=True)
@pytest.fixture
def setup_data():
    config = {
        'format': 'sql:mysql',
        'url': 'mysql://root:password@localhost:3306/menagerie',
        'entities': ('pet', ),
        }

    #print("\nSetting up…")
    yield config
    # Teardown: Clean up resources (if any) after the test
    #print("\nTearing down…")


@pytest.mark.parametrize(
    ('attr', 'expected'),
    [
        ('scheme', 'mysql'),
        ('username', 'root'),
        ('password', 'password'),
        ('hostname', 'localhost'),
        ('port', 3306),
        ('path', '/menagerie'),
    ],
)
def test_smoke_url_format(setup_data, attr, expected):
    from urllib.parse import urlparse
    url = urlparse(setup_data['url'])
    assert getattr(url, attr) == expected


@pytest.mark.skipif(
        (_py.major, _py.minor) < (3, 8),
        reason="Requires python3.8 or higher"
    )
def test_getDatabase_success(setup_data):
    setup_data['entities'] = ('person', 'pet', )
    tree = heimdall.getDatabase(**setup_data)
    assert len(heimdall.getProperties(tree)) == 11
    assert len(heimdall.getEntities(tree)) == 2
    assert len(heimdall.getItems(tree)) == 13


@pytest.mark.skipif(
        (_py.major, _py.minor) < (3, 8),
        reason="Requires python3.8 or higher"
    )
def test_getDatabase_connection_failure(setup_data):
    from mysql.connector.errors import ProgrammingError
    setup_data['url'] = 'mysql://localhost:3306/wtf'
    try:
        tree = heimdall.getDatabase(**setup_data)
        assert False
    except ProgrammingError:
        assert True


@pytest.mark.parametrize(
    ('table', 'row', 'type', 'min', 'max'),
    [
     (
        'test_table',
        ('name', 'varchar(20)', 'utf8mb4_unicode_ci', 'NO', '', None, '',
         'select,insert,update,references', 'Name of this adorable animal'),
        'text', 1, 1
     ),
     (
        'test.table',
        ('sex', 'char(1)', 'utf8mb4_unicode_ci', 'YES', '', None, '',
         'select,insert,update,references', ''),
        'text', 0, 1
     ),
     (
         'blabla',
         ('birth', 'date', None, 'YES', '', None, '',
          'select,insert,update,references', '出生日期'),
         'datetime', 0, 1
     ),
    ],
)
def test_create_attribute(setup_data, table, row, type, min, max):
    from heimdall.connectors.mysql import _create_attribute
    result = _create_attribute(row, table)
    assert result.tag == 'attribute'
    assert len(result.attrib) == 4
    assert result.get('id') == f'{table}.{row[0]}_attr'
    assert result.get('pid') == f'{table}.{row[0]}'
    assert result.min == min
    assert result.max == max
    assert result.type == type
    assert result.name[None] == [row[0], ]
    node = result.description
    assert len(result.description) == 0 or result.description[None] == [row[8], ]


@pytest.mark.parametrize(
    ('table', 'row', 'type'),
    [
     (
        'test_table',
        ('name', 'varchar(20)',
         None, None, None, None, None, None,  # these are unused
         'Name of this adorable animal'),
        'text',
     ),
     (
        'test.table',
        ('sex', 'char(1)',
         None, None, None, None, None, None,  # these are unused
         ''),
        'text',
     ),
     (
         'fzqs',
         ('birth', 'date',
          None, None, None, None, None, None,  # these are unused
          '出生日期'),
         'datetime',
     ),
    ],
)
def test_create_property(setup_data, table, row, type):
    from heimdall.connectors.mysql import _create_property
    result = _create_property(table, row)
    assert result.tag == 'property'
    assert len(result.attrib) == 1
    assert result.get('id') == f'{table}.{row[0]}'
    assert result.type == type
    assert result.name[None] == [row[0], ]
    assert len(result.description) < 1 or result.description[None] == [row[8], ]


@pytest.mark.skipif(
        (_py.major, _py.minor) < (3, 8),
        reason="Requires python3.8 or higher"
    )
def test_create_entity(setup_data):
    from heimdall.connectors.mysql import _create_entity
    from urllib.parse import urlparse
    url = urlparse(setup_data['url'])
    from mysql.connector import connect
    connection = connect(database=url.path.split('/')[1],
                         user=url.username, password=url.password,
                         host=url.hostname, port=url.port)
    with connection.cursor() as cursor:
        for table in setup_data['entities']:

            result, properties = _create_entity(table, cursor)
            assert result.tag == 'entity'
            assert len(result.attrib) == 1
            assert result.get('id') == table
            assert len(result.children) == 9
            assert len(result.children) == len(properties)+2
            tags = ['name', 'description', 'attribute', ]
            for child in result.children:
                assert child.tag in tags
                if child.tag == 'name':
                    assert child.text == table
                elif child.tag == 'attribute':
                    assert len(child.attrib) == 4
                    assert child.get('id').startswith(table)
                    assert child.get('id').endswith('_attr')
                    assert child.get('pid').startswith(table)
                    name = child.get('pid').split('.')[-1]
                    assert child.get('min') is not None
                    assert child.get('max') is not None
                    assert child.type is not None
                    assert child.name[None] == [name, ]
                    d = child.description
                    assert len(d) < 1 or d[None] is not None
    connection.close()


@pytest.mark.skipif(
        (_py.major, _py.minor) < (3, 8),
        reason="Requires python3.8 or higher"
    )
def test_create_tree(setup_data):
    from heimdall.connectors.mysql import _create_tree
    from urllib.parse import urlparse
    url = urlparse(setup_data['url'])
    from mysql.connector import connect
    connection = connect(database=url.path.split('/')[1],
                         user=url.username, password=url.password,
                         host=url.hostname, port=url.port)
    with connection.cursor() as cursor:
        tree = _create_tree(setup_data['entities'], cursor)
        assert len(tree.children) == 3  # items, entities, properties
        properties = heimdall.getProperties(tree)
        assert len(properties) == 7

        entities = heimdall.getEntities(tree)
        assert len(entities) == 1
        attributes = heimdall.getAttributes(entities[0])
        assert len(attributes) == len(properties)

        items = heimdall.getItems(tree)
        assert len(items) == 9
        for item in items:
            metadata = item.metadata
            assert len(metadata) <= 7
            for meta in metadata:
                assert meta.tag == 'metadata'
                assert meta.get('aid') is not None
                assert meta.get('aid') in [a.get('id') for a in attributes]
                assert meta.get('pid') is not None
                assert meta.get('pid') in [p.get('id') for p in properties]

    connection.close()


@pytest.mark.skipif(
        (_py.major, _py.minor) < (3, 8),
        reason="Requires python3.8 or higher"
    )
def test_get_database(setup_data):
    from urllib.parse import urlparse
    url = urlparse(setup_data['url'])
    #from .logger import getLogger
    #log = getLogger(__name__)
    from mysql.connector import connect
    connection = connect(database=url.path.split('/')[1],
                         user=url.username, password=url.password,
                         host=url.hostname, port=url.port)
    with connection.cursor() as cursor:
        for table in setup_data['entities']:

            result = cursor.execute(f'SHOW FULL COLUMNS FROM {table}')
            rows = cursor.fetchall()
            #for row in rows:
            #   log.info(row)
            #log.info('-----')

            result = cursor.execute(f'SELECT * FROM {table}')
            rows = cursor.fetchall()
            #for row in rows:
            #   log.info(row)
    connection.close()


@pytest.mark.parametrize(
    ('given', 'expected'),
    [
        ('date', 'datetime'),
        ('datetime', 'datetime'),
        ('timestamp', 'datetime'),
        ('varchar', 'text'),
        ('varchar(64)', 'text'),
        ('varchar(wtf)', 'text'),
        ('char', 'text'),
        ('char(16)', 'text'),
        ('char(wtf)', 'text'),
        ('charioteer', 'text'),  # :/
        ('tinytext', 'text'),
        ('tinytext(1)', 'text'),
        ('tinytextual', 'text'),  # :/
        ('int', 'number'),
        ('int(16)', 'number'),
        ('int(wtf)', 'number'),
        ('integral', 'number'),  # :/
        ('tinyint', 'number'),
        ('tinyint(1)', 'number'),
        ('tinyintegral', 'number'),  # :/
    ],
)
def test_type_sql2hera(given, expected):
    from heimdall.connectors.mysql import _type_sql2hera
    assert _type_sql2hera(given) == expected


@pytest.mark.parametrize(
    ('given'),
    [
        ('time'),
        ('varcha'),
        ('cha'),
        ('whatever'),
    ],
)
def test_type_sql2hera_unknown(given):
    from heimdall.connectors.mysql import _type_sql2hera
    try:
        _type_sql2hera(given)
        assert False
    except ValueError:
        assert True
